# SPDX-FileCopyrightText: StorPool <support@storpool.com>
# SPDX-License-Identifier: BSD-2-Clause
"""Prepare, extract, and possibly build packages."""

from __future__ import annotations

import argparse
import dataclasses
import json
import pathlib
import subprocess  # noqa: S404
import sys
from typing import TYPE_CHECKING, Callable, Iterable, TypeVar

import utf8_locale

from . import build
from . import checkout
from . import chroot
from . import defs
from . import repo
from . import util
from .scan import deb as s_deb
from .scan import rpm as s_rpm


if TYPE_CHECKING:
    from typing import Final

    Callback = Callable[[defs.Config], None]

    T = TypeVar("T")

    SubParser = argparse._SubParsersAction[argparse.ArgumentParser]  # noqa: SLF001
else:
    SubParser = argparse._SubParsersAction  # noqa: SLF001


@dataclasses.dataclass(frozen=True)
class CommonOpt:
    """A definition for a common command-line option."""

    short: str
    long: str
    help: str
    has_argument: bool = True
    required: bool = True
    default: str | None = None


COMMON_OPTS: Final = {
    "branch": CommonOpt(short="-b", long="--branch", help="the Git branch to operate on"),
    "branch_only": CommonOpt(
        short="-B",
        long="--branch-only",
        help="only show the branch names",
        has_argument=False,
        required=False,
    ),
    "mirror": CommonOpt(
        short="-m",
        long="--mirror",
        help="the Debian mirror/proxy to use",
        required=False,
    ),
    "repo": CommonOpt(
        short="-r",
        long="--repo",
        help="the path to the package repository to update",
    ),
    "sprepo": CommonOpt(
        short="-R",
        long="--sp-repo",
        help="the path to the add-storpool-repo tree",
    ),
    "pkgname": CommonOpt(short="", long="pkgname", help="the package to operate on"),
    "remotepkg": CommonOpt(
        short="-P",
        long="--sp-pkg-remote",
        help="the path to sp-pkg on the remote host",
        required=False,
        default="sp-pkg",
    ),
}


def cmd_clean(cfg: defs.Config) -> None:
    """Clean up the work/ directory."""
    subprocess.check_call(["rm", "-rf", "--", cfg.path.project / "work"], env=cfg.subenv)


def cmd_list_branches(_cfg: defs.Config) -> None:
    """List the branches known to sp-pkg."""
    print("\n".join(f"{branch.name}\t{branch.flavor}" for branch in defs.ORDERED_BRANCHES))


def cmd_list_config(cfg: defs.Config) -> None:
    """List the current configuration."""
    stripped: Final = dataclasses.replace(
        cfg,
        subenv={key: cfg.subenv[key] for key in ("LC_ALL", "LANGUAGE") if key in cfg.subenv},
    )
    print(
        json.dumps(
            {
                "format": {"version": {"major": 0, "minor": 1}},
                "sp_pkg": {
                    "config": {
                        "branch": {
                            "name": stripped.branch.name,
                            "flavor": stripped.branch.flavor,
                            "repo_id": stripped.branch.repo_id,
                        }
                        if stripped.branch is not None
                        else None,
                        "mirror": stripped.mirror,
                        "path": {
                            "home": str(stripped.path.home),
                            "project": str(stripped.path.project),
                            "package": str(stripped.path.package),
                            "dist": str(stripped.path.dist),
                            "repo": str(stripped.path.repo)
                            if stripped.path.repo is not None
                            else None,
                            "sprepo": str(stripped.path.sprepo)
                            if stripped.path.sprepo is not None
                            else None,
                        },
                        "pkgname": stripped.pkgname,
                        "subenv": stripped.subenv,
                        "verbose": stripped.verbose,
                    },
                },
            },
            indent=2,
        ),
    )


def cmd_list_packages(cfg: defs.Config) -> None:
    """List the available packages and branches."""
    print("\n".join(sorted(util.list_packages(cfg))))


def get_projdir(subenv: dict[str, str], workdir: pathlib.Path) -> pathlib.Path:
    """Get the base directory of the Git project checkout."""
    lines: Final = subprocess.check_output(
        ["git", "worktree", "list", "--porcelain"],
        shell=False,
        env=subenv,
        cwd=workdir,
        encoding="UTF-8",
    ).splitlines()
    if not lines:
        sys.exit("Could not determine the current Git checkout directory")
    fields: Final = lines[0].partition("worktree ")
    if len(fields) != 3 or fields[0] or not fields[2]:  # noqa: PLR2004
        sys.exit(f"Weird first line from `git worktree list`: {lines[0]!r}")
    return pathlib.Path(fields[2])


def optional(
    args: argparse.Namespace,
    name: str,
    vtype: type[T],
    default: T | None = None,
) -> T | None:
    """Extract a value if specified."""
    val: Final = getattr(args, name, None)
    if val is None:
        return default

    return vtype(val)  # type: ignore[call-arg]


def optional_def(
    args: argparse.Namespace,
    name: str,
    vtype: type[T],
    default: T,
) -> T:
    """Extract a value if specified."""
    val: Final = optional(args, name, vtype, default)
    assert val is not None  # noqa: S101  # mypy needs this
    return val


def get_branch(args: argparse.Namespace) -> defs.Branch | None:
    """Get the definition of a branch if one is specified."""
    branch_name: Final = optional(args, "branch", str)
    if branch_name is None:
        return None

    branch: Final = defs.BRANCHES.get(branch_name)
    if branch is None:
        sys.exit(
            f"Unknown branch {branch_name!r}, must be one of "
            + ", ".join(branch.name for branch in defs.ORDERED_BRANCHES),
        )
    return branch


def build_config(
    args: argparse.Namespace,
) -> tuple[defs.Config, Callback]:
    """Process the command-line arguments."""
    pkgname: Final = optional(args, "pkgname", str)
    subenv: Final = utf8_locale.get_utf8_env()

    proj_override: Final = getattr(args, "project_dir", None)
    projdir: Final = get_projdir(
        subenv,
        pathlib.Path(proj_override if proj_override is not None else "."),
    )

    branch: Final = get_branch(args)

    paths: Final = defs.Paths(
        home=projdir / defs.REL_HOME,
        project=projdir,
        package=projdir / pkgname if pkgname is not None else projdir,
        work=projdir / "work" / branch.name / pkgname / pkgname
        if branch is not None and pkgname is not None
        else None,
        dist=optional_def(args, "distdir", pathlib.Path, projdir.parent / "dist"),
        repo=optional(args, "repo", pathlib.Path),
        sprepo=optional(args, "sp_repo", pathlib.Path),
        remote_sppkg=optional(args, "sp_pkg_remote", pathlib.Path),
    )

    if paths.sprepo is not None:
        assert branch is not None  # noqa: S101  # mypy needs this
        if not (paths.sprepo / "storpool_variant").is_file():
            sys.exit(f"No storpool_variant at {paths.sprepo}")
        if not (paths.sprepo / branch.repo_id).is_dir():
            sys.exit(f"No {branch.repo_id} in {paths.sprepo}")

    func: Final = getattr(args, "func", None)
    if func is None:
        sys.exit("No command specified")
    return (
        defs.Config(
            branch=branch,
            branch_only=getattr(args, "branch_only", False),
            log=util.build_logger(debug=args.verbose, verbose=True),
            mirror=optional(args, "mirror", str),
            path=paths,
            pkgname=pkgname,
            subenv=subenv,
            verbose=bool(args.verbose),
        ),
        func,
    )


def parse_args() -> tuple[defs.Config, Callback]:
    """Parse the command-line arguments."""

    def add_common(
        parent: SubParser,
        cmd: str,
        *,
        help: str,  # noqa: A002
        func: Callback,
        opts: Iterable[str],
    ) -> argparse.ArgumentParser:
        """Add some common options, e.g. verbose."""
        p_cmd: Final = parent.add_parser(cmd, help=help)
        for optname in opts:
            opt = COMMON_OPTS[optname]
            if opt.has_argument:
                if opt.long.startswith("--"):
                    p_cmd.add_argument(
                        opt.short,
                        opt.long,
                        type=str,
                        required=opt.required,
                        default=opt.default,
                        help=opt.help,
                    )
                else:
                    # A positional argument.
                    p_cmd.add_argument(opt.long, type=str, help=opt.help)
            else:
                p_cmd.add_argument(
                    opt.short,
                    opt.long,
                    action="store_true",
                    required=opt.required,
                    help=opt.help,
                )

        p_cmd.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="verbose operation; display diagnostic output",
        )
        p_cmd.set_defaults(func=func)

        return p_cmd

    parser: Final = argparse.ArgumentParser(prog="sp_pkg")

    parser.add_argument(
        "-C",
        "--project-dir",
        type=str,
        help="specify a different project directory",
    )

    subp: Final = parser.add_subparsers()

    add_common(
        subp,
        "clean",
        help="remove the work directory tree",
        func=cmd_clean,
        opts=[],
    )

    add_common(
        subp,
        "rebuild",
        help="alias for 'pkg rebuild'",
        func=build.cmd_rebuild,
        opts=["branch", "pkgname"],
    )

    p_cmd = subp.add_parser("branch", help="branch actions")
    subsubp = p_cmd.add_subparsers()

    add_common(
        subsubp,
        "list",
        help="list all branches known by sp-pkg",
        func=cmd_list_branches,
        opts=[],
    )

    p_cmd = subp.add_parser("chroot", help="build chroot environment actions")
    subsubp = p_cmd.add_subparsers()

    add_common(
        subsubp,
        "list",
        help="list the available build chroot environments",
        func=chroot.cmd_list_chroots,
        opts=["branch_only"],
    )

    add_common(
        subsubp,
        "remove",
        help="remove a chroot build environment",
        func=chroot.cmd_remove_chroot,
        opts=["branch"],
    )

    add_common(
        subsubp,
        "setup",
        help="set up a chroot build environment",
        func=chroot.cmd_setup_chroot,
        opts=["branch", "mirror", "sprepo"],
    )

    add_common(
        subsubp,
        "update",
        help="update the package cache in a chroot build environment",
        func=chroot.cmd_update_chroot,
        opts=["branch"],
    )

    p_cmd = subp.add_parser("config", help="configuration-related actions")
    subsubp = p_cmd.add_subparsers()

    add_common(
        subsubp,
        "list",
        help="show the current configuration",
        func=cmd_list_config,
        opts=[],
    )

    p_cmd = subp.add_parser("pkg", help="package actions")
    subsubp = p_cmd.add_subparsers()

    add_common(
        subsubp,
        "build",
        help="build a checked-out package",
        func=build.cmd_build,
        opts=["branch", "pkgname"],
    )

    add_common(
        subsubp,
        "checkout",
        help="check a single package out for building",
        func=checkout.cmd_checkout,
        opts=["branch", "pkgname"],
    )

    add_common(
        subsubp,
        "list",
        help="list the available packages and branches",
        func=cmd_list_packages,
        opts=[],
    )

    add_common(
        subsubp,
        "rebuild",
        help="perform a clean/checkout/build cycle for a package",
        func=build.cmd_rebuild,
        opts=["branch", "pkgname"],
    )

    p_cmd = subp.add_parser("remote", help="remote build host actions")
    subsubp = p_cmd.add_subparsers()

    add_common(
        subsubp,
        "rebuild",
        help="rebuild a package on a remote build host",
        func=build.cmd_remote_rebuild,
        opts=["branch", "pkgname", "remotepkg"],
    )

    add_common(
        subsubp,
        "update",
        help="update a chroot on a remote build host",
        func=chroot.cmd_remote_update,
        opts=["branch", "remotepkg"],
    )

    p_cmd = subp.add_parser("repo", help="repository actions")
    subsubp = p_cmd.add_subparsers()

    add_common(
        subsubp,
        "include",
        help="add a package to a repository",
        func=repo.cmd_include,
        opts=["branch", "repo", "pkgname"],
    )

    add_common(
        subsubp,
        "stash",
        help="stash a package into a quasi-repository",
        func=repo.cmd_stash,
        opts=["branch", "repo", "pkgname"],
    )

    return build_config(parser.parse_args())


def main() -> None:
    """Parse options, run commands."""
    try:
        cfg, func = parse_args()
        func(cfg)
    except subprocess.CalledProcessError as err:
        sys.exit(f"An external command failed: {err}")
    except s_rpm.RepoError as err:
        sys.exit(f"RPM/Yum packaging/repository error: {err}")
    except s_deb.RepoError as err:
        sys.exit(f"Debian/Ubuntu packaging/repository error: {err}")


if __name__ == "__main__":
    main()
