# SPDX-FileCopyrightText: StorPool <support@storpool.com>
# SPDX-License-Identifier: BSD-2-Clause
"""Build a checked-out package."""

from __future__ import annotations

import json
import pathlib
import subprocess  # noqa: S404
import sys
import typing

from . import checkout
from . import defs
from . import util
from .scan import deb as s_deb
from .scan import rpm as s_rpm


if typing.TYPE_CHECKING:
    from typing import Final


REMOTE_BASE: Final = pathlib.Path("/var/lib/sp-pkg/sp")


def build_debian(cfg: defs.Config) -> list[pathlib.Path]:
    """Build a Debian package and collect the output."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.pkgname is not None  # noqa: S101  # mypy needs this
    assert cfg.path.work is not None  # noqa: S101  # mypy needs this
    chroot_name: Final = cfg.branch.chroot_name()
    cfg.log.debug(
        "- running sbuild for %(pkgname)s in %(chroot_name)s",
        {"pkgname": cfg.pkgname, "chroot_name": chroot_name},
    )

    workbase: Final = cfg.path.work.parent
    cfg.log.debug("- removing any *.changes files in %(workbase)s", {"workbase": workbase})
    for path in cfg.path.project.glob("*.changes"):
        if path.is_file():
            cfg.log.debug("  - %(path)s", {"path": path})
            path.unlink()

    changelog_file: Final = cfg.path.work / "debian/changelog"
    orig_changelog: Final = changelog_file.read_text(encoding="UTF-8")
    prepend, opt_since_ver = s_deb.get_changelog_prologue(cfg)
    if prepend:
        cfg.log.debug(
            "- prepending %(count)d chars to %(changelog_file)s",
            {"count": len(prepend), "changelog_file": changelog_file},
        )
        changelog_file.write_text(prepend + orig_changelog, encoding="UTF-8")

    if not (cfg.path.home / ".sbuildrc").is_file():
        sys.exit(f"No {cfg.path.home}/.sbuildrc file")
    sbuild_env: Final = dict(cfg.subenv)
    sbuild_env["HOME"] = str(cfg.path.home)

    try:
        sbuild_cmd: Final[defs.PathList] = [
            "sbuild",
            "--no-clean-source",
            "--force-orig-source",
            "-c",
            chroot_name,
        ]
        if opt_since_ver:
            sbuild_cmd.extend(["--debbuildopt", opt_since_ver])
        cfg.check_call(
            sbuild_cmd,
            cwd=cfg.path.work,
            env=sbuild_env,
        )

        cfg.log.debug("- looking for a *.changes file in %(workbase)s", {"workbase": workbase})
        built: Final = s_deb.scan_built(cfg, workbase)
        return sorted([*built.files, built.changes, built.build_log])
    finally:
        if prepend:
            cfg.log.debug(
                "- restoring the original contents of %(changelog_file)s",
                {"changelog_file": changelog_file},
            )
            changelog_file.write_text(orig_changelog, encoding="UTF-8")


def build_centos(cfg: defs.Config) -> list[pathlib.Path]:
    """Run a mock build for a package."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.pkgname is not None  # noqa: S101  # mypy needs this
    assert cfg.path.work is not None  # noqa: S101  # mypy needs this
    vers: Final = cfg.branch.name.split("/", maxsplit=1)[-1]
    chroot_name: Final = cfg.branch.chroot_name()
    cfg.log.debug(
        "Building %(pkgname)s in %(chroot_name)s using mock",
        {"pkgname": cfg.pkgname, "chroot_name": chroot_name},
    )

    srcdir: Final = cfg.path.work / "src"
    spec_file: Final = srcdir / (cfg.pkgname + ".spec")
    cfg.log.debug("- looking for %(spec_file)s", {"spec_file": spec_file})
    if not spec_file.is_file():
        sys.exit(f"No spec file {spec_file}")

    outdir: Final = cfg.path.work / "result"
    cfg.log.debug("- recreating %(outdir)s", {"outdir": outdir})
    cfg.check_call(["rm", "-rf", "--", outdir])
    outdir.mkdir(mode=0o755)

    srpmdir: Final = outdir / "srpm"
    srpmdir.mkdir(mode=0o755)

    cfg.log.debug("- building the source RPM in %(srpmdir)s", {"srpmdir": srpmdir})
    opt_pkg: Final = "--yum" if vers == "7" else "--dnf"
    cfg.check_call(
        [
            "mock",
            "-r",
            chroot_name,
            opt_pkg,
            f"--define=dist .storpool.c{vers}",
            "--buildsrpm",
            "--result",
            srpmdir,
            "--spec",
            spec_file,
            "--source",
            srcdir,
        ],
    )

    cfg.log.debug("- looking for the source RPM in %(srpmdir)s", {"srpmdir": srpmdir})
    built_src: Final = s_rpm.scan_built(cfg, srpmdir)
    cfg.log.debug("  - source package: %(name)s", {"name": built_src.source.path.name})
    cfg.log.debug("  - %(count)d binary packages:", {"count": len(built_src.binaries)})
    for rpmf in built_src.binaries:
        cfg.log.debug("    - %(path_name)s", {"path_name": rpmf.path.name})
    if built_src.binaries:
        sys.exit(
            "Did not expect any binary packages from the source RPM build: "
            + " ".join(sorted(rpmf.path.name for rpmf in built_src.binaries)),
        )

    srpm: Final = built_src.source
    if srpm.name != cfg.pkgname:
        sys.exit(f"Unexpected name for {srpm.path}: expected '{cfg.pkgname}', got '{srpm.name}'")

    rpmdir: Final = outdir / "rpm"
    rpmdir.mkdir(mode=0o755)
    cfg.log.debug("- building the actual package in %(rpmdir)s", {"rpmdir": rpmdir})
    cfg.check_call(
        [
            "mock",
            "-r",
            chroot_name,
            opt_pkg,
            f"--define=dist .storpool.c{vers}",
            "--rebuild",
            "--result",
            rpmdir,
            "--",
            srpm.path,
        ],
    )

    cfg.log.debug("- hopefully there are some files in %(rpmdir)s now", {"rpmdir": rpmdir})
    built: Final = s_rpm.scan_built(cfg, rpmdir)
    cfg.log.debug("  - source package: %(path_name)s", {"path_name": built.source.path.name})
    cfg.log.debug("  - %(count)d binary packages:", {"count": len(built.binaries)})
    for rpmf in built.binaries:
        cfg.log.debug("    - %(name)s", {"name": rpmf.path.name})
    if not built.binaries:
        sys.exit("Expected at least one binary package to be built")

    if built.source.name != cfg.pkgname:
        sys.exit(
            f"Unexpected name for {built.source.path}: "
            f"expected '{cfg.pkgname}', got '{built.source.name}'",
        )

    return sorted(path for path in rpmdir.iterdir() if path.is_file())


def recreate_builddir(cfg: defs.Config, package: defs.Package) -> pathlib.Path:
    """Recreate the directory where the build results will be placed."""
    buildbr: Final = cfg.path.project / "build" / package.branch
    buildbr.mkdir(mode=0o755, parents=True, exist_ok=True)

    builddir: Final = buildbr / package.name
    cfg.log.debug("- recreating %(builddir)s", {"builddir": builddir})
    cfg.check_call(["rm", "-rf", "--", builddir])
    builddir.mkdir(mode=0o755)

    return builddir


def cmd_build(cfg: defs.Config) -> None:
    """Build a checked-out package."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.pkgname is not None  # noqa: S101  # mypy needs this

    cfg.log.debug(
        "Building %(pkgname)s for %(branch_name)s in %(package)s",
        {"pkgname": cfg.pkgname, "branch_name": cfg.branch.name, "package": cfg.path.package},
    )
    package: Final = util.get_package_for_branch(cfg)

    builddir: Final = recreate_builddir(cfg, package)

    files: Final = build_debian(cfg) if cfg.branch.flavor == "debian" else build_centos(cfg)
    missing: Final = [path for path in files if not path.is_file()]
    if missing:
        sys.exit(f"Internal error: missing files: {missing}")
    cfg.log.debug(
        "- moving %(count)d files to %(builddir)s",
        {"count": len(files), "builddir": builddir},
    )
    for src in files:
        dst = builddir / src.name
        cfg.log.debug("  - %(src)s -> %(dst)s", {"src": src, "dst": dst})
        src.rename(dst)


def cmd_rebuild(cfg: defs.Config) -> None:
    """Implement a rebuild-from-scratch cycle."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.pkgname is not None  # noqa: S101  # mypy needs this
    cfg.log.debug(
        "Rebuilding %(pkgname)s for %(branch_name)s",
        {"pkgname": cfg.pkgname, "branch_name": cfg.branch.name},
    )

    checkout.cmd_checkout(cfg)
    cmd_build(cfg)


def get_remote_info(
    cfg: defs.Config,
) -> tuple[str, dict[str, pathlib.Path], dict[str, defs.PathList]]:
    """Build the remote paths, obtain the locale settings to use."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.path.remote_sppkg is not None  # noqa: S101  # mypy needs this
    branch_name: Final = cfg.branch.name

    try:
        rhost: Final = defs.BRANCH_DEFS[branch_name].remote_host
    except KeyError:
        sys.exit(f"Don't know where to build for the {branch_name} branch")
    cfg.log.debug("- will build on %(rhost)s", {"rhost": rhost})

    rpath: Final = {"base": REMOTE_BASE, "pkg": cfg.path.remote_sppkg}
    rpath.update(
        {
            "project": rpath["base"] / cfg.path.project.name,
            "dist": rpath["base"] / "dist",
        },
    )
    if cfg.pkgname is not None:
        rpath.update({"build": rpath["project"] / "build" / cfg.branch.name / cfg.pkgname})

    cfg.log.debug("- get the locale name to use for %(rhost)s", {"rhost": rhost})
    try:
        rcfg: Final = cfg.check_output(
            [
                "env",
                "LC_ALL=C",
                "LANG=",
                "LANGUAGE=",
                "ssh",
                "--",
                rhost,
                rpath["pkg"],
                "-C",
                rpath["project"],
                "config",
                "list",
            ],
        )
    except subprocess.CalledProcessError as err:
        sys.exit(f"Could not ask {rpath['pkg']} for the locale to use on {rhost}: {err}")
    try:
        rcfg_data: Final = json.loads(rcfg)
    except json.JSONDecodeError as err:
        sys.exit(
            f"Could not decode the response from "
            f"`{rpath['pkg']} list config` on {rhost} as JSON: {err}",
        )
    if not isinstance(rcfg_data, dict):
        sys.exit(
            f"`{rpath['pkg']} list config` on {rhost} did not output "
            f"a dictionary but {type(rcfg_data).__name__} instead",
        )
    try:
        rem_lc: Final = rcfg_data["sp_pkg"]["config"]["subenv"]["LC_ALL"]
        rem_lang: Final = rcfg_data["sp_pkg"]["config"]["subenv"]["LANGUAGE"]
    except KeyError:
        sys.exit(f"No LC_ALL or LANGUAGE in the output of `{rpath['pkg']} list config` on {rhost}")
    # Argh, should we try to quote these?!
    remote: Final[dict[str, defs.PathList]] = {
        "ssh": [
            "env",
            "LC_ALL=C",
            "LANG=",
            "LANGUAGE=",
            "ssh",
            "--",
            rhost,
            "env",
            f"LC_ALL={rem_lc}",
            f"LANGUAGE={rem_lang}",
        ],
        "rsync": [
            "rsync",
            "-e",
            "env LC_ALL=C LANG= LANGUAGE= ssh",
        ],
    }
    cfg.log.debug("- remote commands to use:")
    for key in ("ssh", "rsync"):
        cfg.log.debug(
            "  - %(key)s: %(cmd)s",
            {"key": key, "cmd": " ".join(str(item) for item in remote[key])},
        )

    return rhost, rpath, remote


def cmd_remote_rebuild(cfg: defs.Config) -> None:
    """Run a rebuild on a remote host."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.pkgname is not None  # noqa: S101  # mypy needs this
    branch_name: Final = cfg.branch.name
    cfg.log.debug(
        "Preparing for a remote build of %(pkgname)s for %(branch_name)s",
        {"pkgname": cfg.pkgname, "branch_name": branch_name},
    )

    rhost, rpath, remote = get_remote_info(cfg)

    package: Final = util.get_package_for_branch(cfg)
    _, distfiles = checkout.find_source_archives(cfg, package)
    cfg.log.debug("- syncing %(count)d source tarball(s)", {"count": len(distfiles)})
    if any(path.parent != cfg.path.dist for path in distfiles):
        sys.exit(f"The distfiles should all be in {cfg.path.dist}: {distfiles!r}")
    cfg.check_call(
        remote["rsync"]
        + ["-av", "--"]
        + [path.name for path in distfiles]
        + [rhost + ":" + str(rpath["dist"]) + "/"],
        cwd=cfg.path.dist,
    )

    cfg.log.debug("- syncing the project directory itself")
    cfg.check_call(
        remote["rsync"]
        + [
            "-avz",
            "--delete",
            "--exclude=build",
            "--exclude=_data/home",
            "--exclude=work",
            "--",
            f"{cfg.path.project}/",
            f"{rhost}:{rpath['project']}/",
        ],
    )

    cfg.log.debug(
        "- removing %(build_path)s if it exists on %(rhost)s",
        {"build_path": rpath["build"], "rhost": rhost},
    )
    cfg.check_call(remote["ssh"] + ["rm", "-rf", "--", rpath["build"]])

    cfg.log.debug("- now running the remote build")
    cfg.check_call(
        remote["ssh"]
        + [
            rpath["pkg"],
            "-C",
            rpath["project"],
            "rebuild",
        ]
        + (["-v"] if cfg.verbose else [])
        + ["-b", cfg.branch.name, "--", cfg.pkgname],
    )

    builddir: Final = recreate_builddir(cfg, package)
    cfg.log.debug("- looks like the remote build succeeded; copying the files over")
    cfg.check_call(
        remote["rsync"]
        + [
            "-avz",
            "--delete",
            "--copy-links",
            "--",
            rhost + ":" + str(rpath["build"]) + "/",
            str(cfg.path.project / "build" / cfg.branch.name / cfg.pkgname) + "/",
        ],
    )

    cfg.log.debug(
        "- and now there are indeed some files in %(builddir)s, right?",
        {"builddir": builddir},
    )
    if cfg.branch.flavor == "debian":
        dbuilt: Final = s_deb.scan_built(cfg, builddir)
        print(f"- got the {dbuilt.source} source package and {len(dbuilt.files)} files")
    else:
        rbuilt: Final = s_rpm.scan_built(cfg, builddir)
        print(
            f"- got the {rbuilt.source.name} source package and "
            f"{len(rbuilt.binaries)} binary packages",
        )
