# SPDX-FileCopyrightText: StorPool <support@storpool.com>
# SPDX-License-Identifier: BSD-2-Clause
"""Set up and manage build chroot environments."""

from __future__ import annotations

import dataclasses
import pathlib
import subprocess  # noqa: S404
import sys
import typing

from . import build
from . import defs


if typing.TYPE_CHECKING:
    from typing import Final


BEFORE_ALMA: Final = 8

RINGROOTS_CONFIG: Final = pathlib.Path("ringroots.toml")


@dataclasses.dataclass(frozen=True)
class Chroot:
    """Common properties of an existing chroot environment."""

    name: str
    dir_path: pathlib.Path
    cfg_file: pathlib.Path


@dataclasses.dataclass(frozen=True)
class Mock(Chroot):
    """Properties specific to a mock chroot environment."""

    boot_path: pathlib.Path


@dataclasses.dataclass(frozen=True)
class SChroot(Chroot):
    """Properties specific to a sbuild schroot environment."""


MOCK_DIR: Final = pathlib.Path("/etc/mock")


def find_mock(cfg: defs.Config) -> list[tuple[str, str]]:
    """List the available sbuild chroots."""
    cfg.log.debug("Listing mock chroots")
    if not MOCK_DIR.is_dir():
        cfg.log.debug("- nothing here")
        return []

    res: Final = []
    for branch in defs.ORDERED_BRANCHES:
        if branch.flavor != "centos":
            continue

        chroot_name = branch.chroot_name()
        cfile = MOCK_DIR / (chroot_name + ".cfg")
        if not cfile.is_file():
            continue

        res.append((branch.name, chroot_name))

    return res


def find_schroots(cfg: defs.Config) -> list[tuple[str, str]]:
    """List the available sbuild chroots."""
    cfg.log.debug("Listing schroots")
    try:
        proc: Final = cfg.Popen(["schroot", "-l"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        assert proc.stdout is not None  # noqa: S101  # mypy needs this
        assert proc.stderr is not None  # noqa: S101  # mypy needs this
        chroots: Final = proc.stdout.read().splitlines()
    except (OSError, subprocess.CalledProcessError) as err:
        cfg.log.debug("Could not run `schroot -l`: %(err)s", {"err": err})
        return []

    res: Final = []
    for branch in defs.ORDERED_BRANCHES:
        if branch.flavor != "debian":
            continue

        chroot_name = branch.chroot_name()
        src_name = f"chroot:{chroot_name}"
        bin_name = f"source:{chroot_name}"
        cfg.log.debug(
            "- check for %(src_name)r %(bin_name)r",
            {"src_name": src_name, "bin_name": bin_name},
        )

        if src_name in chroots:
            if bin_name not in chroots:
                sys.exit(f"Weird: found {src_name} but not {bin_name}")

            res.append((branch.name, chroot_name))
        elif bin_name in chroots:
            sys.exit(f"Weird: found {bin_name} but not {src_name}")

    return res


def cmd_list_chroots(cfg: defs.Config) -> None:
    """List the available build chroot environments."""
    for branch_name, chroot_name in find_schroots(cfg) + find_mock(cfg):
        print("\t".join([branch_name] + ([] if cfg.branch_only else [chroot_name])))


def get_mock(cfg: defs.Config) -> Mock | None:
    """Get the properties of a mock chroot if it exists."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    chroot_name: Final = cfg.branch.chroot_name()

    cfg_file: Final = MOCK_DIR / (chroot_name + ".cfg")
    if not cfg_file.is_file():
        return None

    proc: Final = cfg.Popen(
        ["mock", "--print-root-path", "-r", chroot_name],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    p_out, _ = proc.communicate()
    res: Final = proc.wait()
    if res != 0:
        return None

    dir_paths: Final = p_out.splitlines()
    if len(dir_paths) != 1:
        cfg.log.debug("%(dir_paths)r", {"dir_paths": dir_paths})
        sys.exit(f"Weird: `mock --print-root-path -r {chroot_name!r}` should output a single line")
    dir_path: Final = pathlib.Path(dir_paths[0])
    boot_path: Final = (
        dir_path.parent.parent / (dir_path.parent.name + "-bootstrap") / dir_path.name
    )

    return Mock(
        name=chroot_name,
        dir_path=dir_path,
        cfg_file=cfg_file,
        boot_path=boot_path,
    )


def get_schroot(cfg: defs.Config) -> SChroot | None:
    """Get the properties of a schroot if it exists."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    chroot_name: Final = cfg.branch.chroot_name()
    proc: Final = cfg.Popen(
        ["schroot", "--config", "-c", chroot_name],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    p_out, _ = proc.communicate()
    res: Final = proc.wait()
    if res != 0:
        return None

    dir_paths: Final = [
        parts[2]
        for parts in (line.partition("directory=") for line in p_out.splitlines())
        if len(parts) == 3 and not parts[0] and parts[1]  # noqa: PLR2004
    ]
    if len(dir_paths) != 1:
        cfg.log.debug("%(dir_paths)r", {"dir_paths": dir_paths})
        sys.exit(
            f"Weird: `schroot --config -c {chroot_name!r}` should "
            f"output a single 'directory=' line",
        )
    dir_path: Final = pathlib.Path(dir_paths[0])

    def get_chroot_cfgfile() -> pathlib.Path:
        """Find the schroot config file."""
        chroot_dir: Final = pathlib.Path("/etc/schroot/chroot.d")
        for cfgfile in chroot_dir.iterdir():
            if not cfgfile.is_file():
                continue

            if f"[{chroot_name}]" in cfgfile.read_text(encoding="UTF-8").splitlines():
                return cfgfile

        sys.exit(f"Could not find a definition for [{chroot_name}] in {chroot_dir}")

    return SChroot(name=chroot_name, dir_path=dir_path, cfg_file=get_chroot_cfgfile())


def schroot_mk_sbuild(cfg: defs.Config, branch_name: str, chroot_name: str) -> SChroot:
    """Run mk-sbuild to create a chroot environment."""
    assert cfg.mirror is not None  # noqa: S101  # mypy needs this
    cfg.log.debug("- creating the %(chroot_name)s sbuild chroot", {"chroot_name": chroot_name})
    cfg.check_call(
        [
            "mk-sbuild",
            "--name",
            chroot_name.replace("-amd64", ""),
            "--debootstrap-mirror",
            cfg.mirror,
            branch_name.split("/", maxsplit=1)[-1],
        ],
    )
    cfg.log.debug("- querying the just-created chroot environment")
    schroot: Final = get_schroot(cfg)
    if schroot is None:
        sys.exit(f"Could not query the {chroot_name} schroot after mk-sbuild")

    lines: Final = schroot.cfg_file.read_text(encoding="UTF-8").splitlines()
    if len([line for line in lines if line == f"[{chroot_name}]"]) != 1:
        sys.exit(f"Could not find a '[{chroot_name}]' line in {schroot.cfg_file}")

    indices_source: Final = [
        idx for idx, line in enumerate(lines) if line.startswith("#source-root-groups=")
    ]
    if len(indices_source) != 1:
        sys.exit(
            f"Expected exactly one line starting with "
            f"'#source-root-groups=' in {schroot.cfg_file}",
        )
    cfg.log.debug("- found a groups line: %(groups)s", {"groups": lines[indices_source[0]]})
    lines[indices_source[0]] = lines[indices_source[0]][1:]

    indices: Final = [idx for idx, line in enumerate(lines) if line == "union-type=aufs"]
    if len(indices) == 1:
        cfg.log.debug("- found an aufs line: %(aufs)s", {"aufs": lines[indices[0]]})
        lines[indices[0]] = lines[indices[0]].replace("aufs", "overlay")
    elif len(indices) != 0:
        sys.exit(f"Expected at most one 'union-type=aufs' line in {schroot.cfg_file}")

    schroot.cfg_file.write_text("".join(line + "\n" for line in lines), encoding="UTF-8")

    # Let us hope we don't have to do this again in the future...
    if branch_name == "debian/bullseye":
        cfg.log.debug("Fixing up the bullseye-security mixup")
        sfile: Final = schroot.dir_path / "etc/apt/sources.list"
        cfg.log.debug("- reading %(sfile)s", {"sfile": sfile})
        current: Final = sfile.read_text(encoding="UTF-8")
        repl_bad: Final = "bullseye/updates"
        repl_good: Final = "bullseye-security"
        updated: Final = current.replace(repl_bad, repl_good)
        if updated == current:
            if repl_good not in current:
                sys.exit(
                    f"Could not replace {repl_bad!r} with {repl_good!r} in {sfile}:\n{current}",
                )
            cfg.log.debug("- seems that it has already been fixed upstream")
        else:
            cfg.log.debug("- writing it back")
            sfile.write_text(updated, encoding="UTF-8")

    return schroot


def schroot_apt_upgrade(cfg: defs.Config, sccmd: list[str]) -> None:
    """Upgrade the Debian installation within the chroot environment."""
    cfg.log.debug("- updating the Debian package information")
    cfg.check_call([*sccmd, "apt-get", "update"])

    cfg.log.debug("- updating the Debian packages")
    cfg.check_call(
        [
            *sccmd,
            "env",
            "DEBIAN_FRONTEND=noninteractive",
            "apt-get",
            "-q",
            "-y",
            "dist-upgrade",
        ],
    )


def schroot_add_storpool_repo(cfg: defs.Config, mount_path: pathlib.Path, sccmd: list[str]) -> None:
    """Copy add-storpool-repo in and run it."""
    cfg.check_call(["rsync", "-a", "--delete", f"{cfg.path.sprepo}/", f"{mount_path}/sp/"])
    cfg.check_call([*sccmd, "/sp/add-storpool-repo.sh"])
    cfg.check_call([*sccmd, "/sp/add-storpool-repo.sh", "-t", "staging"])
    cfg.check_call([*sccmd, "/sp/add-storpool-repo.sh", "-t", "infra"])


def schroot_activate_backports(cfg: defs.Config, mount_path: pathlib.Path) -> None:
    """Activate the backports repository if needed."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    cfg.log.debug("- activating the backports repository if needed")
    if cfg.branch.name == "debian/unstable":
        cfg.log.debug("  - not needed")
        return

    suite: Final = cfg.branch.name.split("/", maxsplit=1)[-1]
    srcfile: Final = mount_path / "etc/apt/sources.list"
    lines: Final = srcfile.read_text(encoding="UTF-8").splitlines()

    backports_lines: Final = [
        " ".join(found)
        for found in (
            fields[:2] + [suite + "-backports"] + fields[3:]
            for fields in (line.split() for line in lines)
            if fields[2] == suite
        )
    ]
    cfg.log.debug("  - adding %(backports_lines)r", {"backports_lines": backports_lines})
    srcfile.write_text(
        "".join(line + "\n" for line in lines + backports_lines),
        encoding="UTF-8",
    )

    pref: Final = mount_path / "etc/apt/preferences.d/debhelper.pref"
    cfg.log.debug("  - creating %(pref)s", {"pref": pref})
    pref.write_text(
        f"""
Package: debhelper
Pin: release a={suite}-backports
Pin-Priority: 990

Package: libdebhelper-perl
Pin: release a={suite}-backports
Pin-Priority: 990

Package: dh-autoreconf
Pin: release a={suite}-backports
Pin-Priority: 990

Package: dwz
Pin: release a={suite}-backports
Pin-Priority: 990
""",
        encoding="UTF-8",
    )


def schroot_trivial_checks(cfg: defs.Config, sccmd: list[str]) -> None:
    """Run a couple of trivial commands to check stuff."""
    cfg.check_call([*sccmd, "apt-cache", "policy", "debhelper"])
    cfg.check_call([*sccmd, "apt-cache", "policy", "sp-python3"])


def setup_schroot(cfg: defs.Config) -> None:
    """Set up a sbuild schroot environment."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.path.sprepo is not None  # noqa: S101  # mypy needs this

    if cfg.mirror is None:
        sys.exit("The --mirror option is required for Debian/Ubuntu chroot environments")
    branch_name: Final = cfg.branch.name
    chroot_name: Final = cfg.branch.chroot_name()
    cfg.log.debug(
        "Set up the %(chroot_name)s sbuild environment for %(branch_name)s",
        {"chroot_name": chroot_name, "branch_name": branch_name},
    )

    cfg.log.debug("- making sure it does not exist already")
    existing: Final = get_schroot(cfg)
    if existing is not None:
        sys.exit(
            f"The {chroot_name} sbuild schroot environment already exists at {existing.dir_path}",
        )

    schroot_mk_sbuild(cfg, branch_name, chroot_name)

    cfg.log.debug("- creating a new session for %(chroot_name)s", {"chroot_name": chroot_name})
    lines: Final = cfg.check_output(
        [
            "schroot",
            "-b",
            "-c",
            "source:" + chroot_name,
            "-u",
            "root",
            "-d",
            "/",
        ],
    ).splitlines()
    if len(lines) != 1:
        sys.exit(f"Unexpected output from `scroot -b -c {chroot_name}`: {lines!r}")
    sess_id: Final = lines[0]
    cfg.log.debug("- got schroot session ID %(sess_id)s", {"sess_id": sess_id})

    try:

        def get_mount_path() -> pathlib.Path:
            """Get the temporary path where the schroot session is mounted."""
            for pfx in (
                "/run/schroot/mount",
                "/var/run/schroot/mount",
                "/var/lib/schroot/mount",
            ):
                path = pathlib.Path(pfx) / sess_id
                if path.is_dir():
                    return path

            sys.exit(f"Could not find the temporary mount path for {sess_id}")

        mount_path: Final = get_mount_path()
        sccmd: Final = ["schroot", "-r", "-c", sess_id, "-u", "root", "-d", "/", "--"]

        schroot_apt_upgrade(cfg, sccmd)
        schroot_add_storpool_repo(cfg, mount_path, sccmd)
        schroot_activate_backports(cfg, mount_path)
        schroot_apt_upgrade(cfg, sccmd)

        schroot_trivial_checks(cfg, sccmd)

        # All done!
        print(f"{branch_name}\t{chroot_name}")
    finally:
        cfg.log.debug("- cleaning up the %(sess_id)s schroot session", {"sess_id": sess_id})
        cfg.check_call(["schroot", "-e", "-c", sess_id])


def setup_mock(cfg: defs.Config) -> None:
    """Set up a mock chroot environment."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.path.sprepo is not None  # noqa: S101  # mypy needs this

    if cfg.mirror is not None:
        raise NotImplementedError("No --mirror for CentOS chroots yet")

    branch_name: Final = cfg.branch.name
    chroot_name: Final = cfg.branch.chroot_name()
    cfg.log.debug(
        "Setting up the %(chroot_name)s mock chroot for %(branch_name)s",
        {"chroot_name": chroot_name, "branch_name": branch_name},
    )

    vers: Final = branch_name.split("/", maxsplit=1)[-1]
    vers_int: Final = int(vers)
    base_name: Final = (
        f"epel-{vers}-x86_64" if vers_int < BEFORE_ALMA else f"alma+epel-{vers}-x86_64"
    )
    base_file: Final = (MOCK_DIR / base_name).with_suffix(".cfg")
    if not base_file.is_file():
        sys.exit(f"Could not find {base_file}")
    base_text: Final = base_file.read_text(encoding="UTF-8")

    cfg_file: Final = MOCK_DIR / (chroot_name + ".cfg")
    cfg_text = base_text

    def check_replace(orig_text: str, r_from: str, r_to: str) -> str:
        """Make sure a replacement has changed something."""
        new_text: Final = orig_text.replace(r_from, r_to)
        if new_text == orig_text:
            sys.exit(f"Could not find {r_from!r} in {base_file}")
        return new_text

    config_root_lines: Final = [
        line for line in cfg_text.splitlines() if line.startswith("config_opts['root'] = ")
    ]
    if len(config_root_lines) != 1:
        sys.exit(f"Expected exactly one config_opts['root'] line in {base_file}")
    cfg_text = check_replace(
        cfg_text,
        config_root_lines[0],
        f"config_opts['root'] = '{chroot_name}'",
    )

    if "config_opts['dist']" in cfg_text:
        cfg_text = check_replace(
            cfg_text,
            f"config_opts['dist'] = 'el{vers}'",
            f"config_opts['dist'] = 'storpool.c{vers}'",
        )

    cfg_text += '''
config_opts['yum.conf'] += """

'''

    for fname in (
        "storpool-centos.repo",
        "storpool-centos-staging.repo",
        "storpool-centos-infra.repo",
    ):
        cfg.log.debug("- adding the repo configuration from %(fname)s", {"fname": fname})
        snippet = (cfg.path.sprepo / cfg.branch.repo_id / fname).read_text(encoding="UTF-8")
        snippet = check_replace(snippet, "gpgcheck=1", "gpgcheck=0")
        snippet = check_replace(snippet, "enabled=0", "enabled=1")
        cfg_text += snippet + "\n"

    cfg_text += '''
"""
'''
    cfg.log.debug(
        "- writing %(count)d chars to %(cfg_file)s",
        {"count": len(cfg_text), "cfg_file": cfg_file},
    )
    cfg_file.write_text(cfg_text, encoding="UTF-8")

    cfg.log.debug("- running `mock --init`")
    cfg.check_call(["mock", "--init", "-r", chroot_name])

    cfg.log.debug("- could it be that we are done here?")


def cmd_setup_chroot(cfg: defs.Config) -> None:
    """Set up a chroot build environment."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    if cfg.branch.flavor == "debian":
        setup_schroot(cfg)
    else:
        setup_mock(cfg)


def cmd_remove_chroot(cfg: defs.Config) -> None:
    """Remove a chroot build environment."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    branch_name: Final = cfg.branch.name
    chroot_name: Final = cfg.branch.chroot_name()
    cfg.log.debug(
        "Trying to remove the %(chroot_name)s chroot environment for %(branch_name)s",
        {"chroot_name": chroot_name, "branch_name": branch_name},
    )
    chroot: Final[Chroot | None] = (
        get_schroot(cfg) if cfg.branch.flavor == "debian" else get_mock(cfg)
    )
    if chroot is None:
        sys.exit(f"No {chroot_name} chroot for {branch_name}")

    if isinstance(chroot, Mock):
        cfg.log.debug("- running `mock scrub all`")
        cfg.check_call(["mock", "--scrub", "all", "-r", chroot_name])

    cfg.log.debug("- the config file %(cfg_file)s", {"cfg_file": chroot.cfg_file})
    chroot.cfg_file.unlink()

    cfg.log.debug("- the chroot directory %(dir_path)s", {"dir_path": chroot.dir_path})
    cfg.check_call(["rm", "-rf", "--", chroot.dir_path])

    if isinstance(chroot, Mock):
        cfg.log.debug("- the bootstrap directory %(boot_path)s", {"boot_path": chroot.boot_path})
        cfg.check_call(["rm", "-rf", "--", chroot.boot_path])


def update_schroot(cfg: defs.Config) -> None:
    """Update the package cache in a sbuild schroot."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    chroot_name: Final = cfg.branch.chroot_name()
    cfg.log.debug(
        "Update the %(chroot_name)s schroot for %(branch_name)s",
        {"chroot_name": chroot_name, "branch_name": cfg.branch.name},
    )

    sess_lines: Final = cfg.check_output(
        [
            "schroot",
            "-b",
            "-c",
            "source:" + chroot_name,
            "-u",
            "root",
            "-d",
            "/",
        ],
    ).splitlines()
    if len(sess_lines) != 1 or not sess_lines[0]:
        sys.exit(f"`schroot -b -c {chroot_name}` output weird {sess_lines!r}")
    sess_id: Final = sess_lines[0]
    cfg.log.debug("- started source session %(sess_id)s", {"sess_id": sess_id})

    try:
        sccmd: Final = ["schroot", "-r", "-c", sess_id, "-u", "root", "-d", "/", "--"]
        cfg.log.debug("- running `apt-get update`")
        cfg.check_call([*sccmd, "apt-get", "-y", "update"])

        cfg.log.debug("- running `apt-get dist-upgrade`")
        cfg.check_call(
            [
                *sccmd,
                "env",
                "DEBIAN_FRONTEND=noninteractive",
                "apt-get",
                "-y",
                "dist-upgrade",
            ],
        )

        cfg.log.debug("- running `apt-get clean`")
        cfg.check_call([*sccmd, "apt-get", "-y", "clean"])
    finally:
        cfg.log.debug("- cleaning up the %(sess_id)s session", {"sess_id": sess_id})
        cfg.check_call(["schroot", "-e", "-c", sess_id])


def update_mock(cfg: defs.Config) -> None:
    """Update the package cache for a mock chroot."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    chroot_name: Final = cfg.branch.chroot_name()
    cfg.log.debug(
        "Recreating the cache and root fs for %(chroot_name)s",
        {"chroot_name": chroot_name},
    )

    cfg.log.debug("- running `mock scrub all`")
    cfg.check_call(["mock", "--scrub", "all", "-r", chroot_name])

    cfg.log.debug("- running `mock init`")
    cfg.check_call(["mock", "--init", "-r", chroot_name])


def cmd_update_chroot(cfg: defs.Config) -> None:
    """Update the package cache in a chroot."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    if cfg.branch.flavor == "debian":
        update_schroot(cfg)
    else:
        update_mock(cfg)


def cmd_remote_update(cfg: defs.Config) -> None:
    """Update the package cache in a chroot on a remote build host."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    branch_name: Final = cfg.branch.name
    cfg.log.debug(
        "Preparing for a remote update of the %(branch_name)s chroot",
        {"branch_name": branch_name},
    )

    _, rpath, remote = build.get_remote_info(cfg)

    cfg.log.debug("- now running the remote update")
    cfg.check_call(
        remote["ssh"]
        + [
            rpath["pkg"],
            "-C",
            rpath["project"],
            "chroot",
            "update",
        ]
        + (["-v"] if cfg.verbose else [])
        + ["-b", cfg.branch.name],
    )
