# SPDX-FileCopyrightText: StorPool <support@storpool.com>
# SPDX-License-Identifier: BSD-2-Clause
"""Remove a source package and any binaries built from it."""

from __future__ import annotations

import argparse
import dataclasses
import pathlib
import sys
import typing

import utf8_locale

from . import defs
from . import repo
from . import util
from .scan import rpm as s_rpm
from .scan import yum as s_yum


if typing.TYPE_CHECKING:
    from typing import Final


@dataclasses.dataclass(frozen=True)
class Config:
    """Runtime configuration for the replacerpm tool."""

    noop: bool
    repo: s_yum.Repo
    pkgdir: pathlib.Path
    verbose: bool


def parse_arguments() -> Config:
    """Parse the command-line arguments."""
    parser: Final = argparse.ArgumentParser(prog="replacerpm")
    parser.add_argument(
        "--noop",
        "-N",
        action="store_true",
        default=False,
        help="no-operation mode",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        help="verbose operation; display diagnostic output",
    )
    parser.add_argument("repo", type=str, help="path to the repository")
    parser.add_argument("pkgdir", type=str, help="path to the new packages")
    args: Final = parser.parse_args()
    return Config(
        noop=args.noop,
        repo=s_yum.Repo(pathlib.Path(args.repo)),
        pkgdir=pathlib.Path(args.pkgdir),
        verbose=args.verbose,
    )


def main() -> None:
    """Parse options, scan packages, run commands."""
    cfg: Final = parse_arguments()

    cwd: Final = pathlib.Path.cwd()
    scfg: Final = defs.Config(
        branch=None,
        branch_only=False,
        log=util.build_logger(debug=cfg.verbose, verbose=True),
        mirror=None,
        path=defs.Paths(
            home=cwd,
            project=cwd,
            package=cwd,
            work=None,
            dist=cwd,
            repo=cwd,
            sprepo=cwd,
            remote_sppkg=None,
        ),
        pkgname=None,
        subenv=utf8_locale.get_utf8_env(),
        verbose=cfg.verbose,
    )

    built: Final = s_rpm.scan_built(scfg, cfg.pkgdir)
    unsigned: Final = s_rpm.verify_sigs(scfg, built)
    if unsigned:
        sys.exit("Unsigned files:\n" + "\n".join(str(path) for path in unsigned))

    cmds_real: Final = repo.examine_rpm(cfg.repo, built)
    cmds: Final = [["echo", *cmd] for cmd in cmds_real] if cfg.noop else cmds_real

    for cmd in cmds:
        scfg.check_call(cmd)


if __name__ == "__main__":
    main()
