# SPDX-FileCopyrightText: StorPool <support@storpool.com>
# SPDX-License-Identifier: BSD-2-Clause
"""Scan a directory for a built Debian package."""

from __future__ import annotations

import dataclasses
import pathlib
import typing

from . import common


if typing.TYPE_CHECKING:
    from typing import Final

    from sp_pkg import defs


try:
    from debian import changelog
    from debian import deb822

    HAVE_DEB_PYTHON = True
except ImportError:
    HAVE_DEB_PYTHON = False


class RepoError(Exception):
    """An error that occurred during the processing of a Debian package."""


@dataclasses.dataclass(frozen=True)
class BranchDef:
    """A definition for a Debian or Ubuntu release branch."""

    name: str
    slug: str


@dataclasses.dataclass(frozen=True)
class DebBuilt(common.Built):
    """A Debian package built in a directory."""

    changes: pathlib.Path
    source: str
    files: list[pathlib.Path]
    build_log: pathlib.Path


BACKPORT_SLUGS: Final = {
    "debian/bookworm": BranchDef(name="Debian 12.x (bookworm)", slug="~bpo12+1"),
    "debian/bullseye": BranchDef(name="Debian 11.x (bullseye)", slug="~bpo11+1"),
    "ubuntu/noble": BranchDef(name="Ubuntu 24.04 (Noble Numbat)", slug="~ubuntu1~24.04.1"),
    "ubuntu/lunar": BranchDef(name="Ubuntu 23.04 (Lunar Lobster)", slug="~ubuntu1~23.04.1"),
    "ubuntu/jammy": BranchDef(name="Ubuntu 22.04 (Jammy Jellyfish)", slug="~ubuntu1~22.04.1"),
    "ubuntu/impish": BranchDef(name="Ubuntu 21.10 (Impish Indri)", slug="~ubuntu1~21.10.1"),
    "ubuntu/focal": BranchDef(name="Ubuntu 20.04 (Focal Fossa)", slug="~ubuntu1~20.04.1"),
    "ubuntu/bionic": BranchDef(name="Ubuntu 18.04 (Bionic Beaver)", slug="~ubuntu1~18.04.1"),
}


def ensure_deb() -> None:
    """Make sure the debian.* modules have been found."""
    if not HAVE_DEB_PYTHON:
        raise RepoError("Could not import the debian.* Python modules")


def scan_built(cfg: defs.Config, path: pathlib.Path) -> DebBuilt:
    """Scan a directory for a built Debian package."""
    cfg.log.debug("- looking for a *.changes file in %(path)s", {"path": path})
    ensure_deb()

    changes_files: Final = list(path.glob("*.changes"))
    if len(changes_files) != 1:
        raise RepoError(f"Expected a single *.changes file in {path}")
    changes_file: Final = changes_files[0]
    if not changes_file.is_file():
        raise RepoError(f"Not a regular file: {changes_file}")

    cfg.log.debug("- looking for a *.build file next to %(changes)s", {"changes": changes_file})
    build_log: Final = changes_file.with_suffix(".build")
    if not build_log.is_file():
        raise RepoError(f"Found {changes_file} but no {build_log}")

    changes: Final = deb822.Changes(changes_file.open(encoding="UTF-8"))
    srcname: Final = changes["Source"]
    assert isinstance(srcname, str)  # noqa: S101  # mypy needs this
    if srcname != cfg.pkgname:
        raise RepoError(f"Unexpected 'Source' in {changes_file}: {srcname}")

    files: Final = []
    cfiles: Final = changes["Files"]
    assert isinstance(cfiles, list)  # noqa: S101  # mypy needs this
    for debf in cfiles:
        assert isinstance(debf, deb822.Deb822Dict)  # noqa: S101  # mypy needs this
        fname = debf["name"]
        assert isinstance(fname, str)  # noqa: S101  # mypy needs this
        files.append(path / fname)

    version: Final = changes["Version"]
    assert isinstance(version, str)  # noqa: S101  # mypy needs this
    return DebBuilt(
        path=path,
        name=cfg.pkgname,
        version=version,
        changes=changes_file,
        source=srcname,
        files=files,
        build_log=build_log,
    )


def get_changelog_prologue(
    cfg: defs.Config,
) -> tuple[str, str | None]:
    """Optionally prepend a "Rebuild for..." changelog entry."""
    ensure_deb()

    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    branch_name: Final = cfg.branch.name
    if branch_name == "debian/unstable":
        return "", None
    slug: Final = BACKPORT_SLUGS.get(branch_name)
    if not slug:
        raise RepoError(f"Unknown Debian-like branch {branch_name!r}")

    chfile: Final = pathlib.Path(cfg.path.package) / "debian/changelog"
    if not chfile.is_file():
        raise RepoError(f"No Debian changelog file {chfile}")
    chlog: Final = changelog.Changelog(chfile.open(encoding="UTF-8"))
    versions: Final = chlog.get_versions()
    if not versions:
        raise RepoError(f"Could not parse any versions in {chfile}")

    bp_ver: Final = chlog.full_version + slug.slug
    current: Final = chlog[0]
    chlog.new_block(
        package=current.package,
        version=bp_ver,
        distributions=branch_name.split("/", maxsplit=1)[-1],
        urgency=current.urgency,
        urgency_comment=current.urgency_comment,
        changes=["", f"  * Rebuild for {slug.name}.", ""],
        author=current.author,
        date=current.date,
        other_pairs=current.other_pairs,
        encoding="UTF-8",
    )
    blk: Final = chlog[0]

    previous: Final = versions[1].full_version if len(versions) > 1 else "0~0~"

    return str(blk), "-v" + previous
