# SPDX-FileCopyrightText: StorPool <support@storpool.com>
# SPDX-License-Identifier: BSD-2-Clause
"""Scan a directory for RPM files."""

from __future__ import annotations

import dataclasses
import re
import sys
import typing

from . import common


if typing.TYPE_CHECKING:
    import pathlib
    from collections.abc import Iterable
    from typing import Final

    from sp_pkg import defs


RE_SP_REPO_KEY: Final = re.compile(
    r""" ^
    (?: gpg [(] )?
    StorPool
    .*
    repo@storpool.com
    """,
    re.X,
)


class RepoError(Exception):
    """An error that occurred while examining the repository."""


@dataclasses.dataclass(frozen=True)
class RPMFile:
    """A single RPM package."""

    path: pathlib.Path
    name: str
    evr: str
    epoch: str
    version: str
    release: str
    arch: str


@dataclasses.dataclass(frozen=True)
class RPMSource(RPMFile):
    """A single RPM source package."""


@dataclasses.dataclass(frozen=True)
class RPMBin(RPMFile):
    """A single RPM binary package."""

    source: str


@dataclasses.dataclass(frozen=True)
class RPMBuilt(common.Built):
    """A set of built packages."""

    source: RPMSource
    binaries: list[RPMBin]


def query_rpm_file(cfg: defs.Config, path: pathlib.Path) -> RPMFile:
    """Query an RPM file for its attributes."""
    lines: Final = cfg.check_output(
        [
            "rpm",
            "-qp",
            "--qf",
            "\\t".join(
                f"%{{{field}}}"
                for field in (
                    "Name",
                    "EVR",
                    "Epoch",
                    "Version",
                    "Release",
                    "Arch",
                    "SourceRPM",
                )
            )
            + "\\n",
            "--",
            path,
        ],
    ).splitlines()
    if len(lines) != 1:
        sys.exit(f"Expected `rpm -qp` to return a single line: {lines!r}")
    fields: Final = lines[0].split("\t")
    if len(fields) != 7:  # noqa: PLR2004
        sys.exit(f"Expected `rpm -qp` to return seven fields: {fields!r}")

    if path.name.endswith(".src.rpm"):
        return RPMSource(
            path=path,
            name=fields[0],
            evr=fields[1],
            epoch=fields[2],
            version=fields[3],
            release=fields[4],
            arch=fields[5],
        )

    return RPMBin(
        path=path,
        name=fields[0],
        evr=fields[1],
        epoch=fields[2],
        version=fields[3],
        release=fields[4],
        arch=fields[5],
        source=fields[6],
    )


def scan_built(cfg: defs.Config, path: pathlib.Path) -> RPMBuilt:
    """Scan a directory for a source package and some binaries."""
    sources: Final[list[RPMSource]] = []
    binaries: Final = []
    for rpmf in path.glob("*.rpm"):
        if not rpmf.is_file():
            raise RepoError(f"Not a regular file: {rpmf}")

        data = query_rpm_file(cfg, rpmf)
        if isinstance(data, RPMSource):
            if sources:
                raise RepoError(
                    f"More than one source RPM package in {path}: "
                    f"first {sources[0].path} and now {rpmf}",
                )

            sources.append(data)
        else:
            assert isinstance(data, RPMBin)  # noqa: S101  # mypy needs this
            binaries.append(data)

    if not sources:
        raise RepoError(f"No RPM source package found in {path}")
    source: Final = sources[0]

    for rpmb in binaries:
        if rpmb.source != source.path.name:
            raise RepoError(
                f"The {rpmb.path} package seems to have been built from "
                f"{rpmb.source} and not {source.path.name}",
            )

    return RPMBuilt(
        path=path,
        name=source.name,
        version=source.evr,
        source=source,
        binaries=binaries,
    )


def find_sp_key(cfg: defs.Config) -> str:
    """Figure out which of the installed RPM keys is ours."""
    cfg.log.debug("- looking for the StorPool repository signing key RPM package")
    for line in cfg.check_output(
        [
            "rpm",
            "-qa",
            "--qf",
            "%{Name}\\t%{Version}\\t%{Summary}\\n",
            "gpg-pubkey",
        ],
    ).splitlines():
        fields = line.split("\t")
        cfg.log.debug("  - examining %(fields)s", {"fields": fields})
        if RE_SP_REPO_KEY.match(fields[2]):
            cfg.log.debug("  - found it!")
            return fields[1]

    raise RepoError("Could not determine the StorPool repository signing key")


def verify_sigs(
    cfg: defs.Config,
    built: RPMBuilt,
    *,
    key_id: str | None = None,
) -> list[pathlib.Path]:
    """Make sure the files are signed."""
    if key_id is None:
        key_id = find_sp_key(cfg)

    re_sig: Final = re.compile(
        (
            rf"V4 \s RSA/SHA (?: 1 | 512 ) \s Signature, \s "
            rf"key \s ID \s {re.escape(key_id)}: \s OK"
        ),
        re.X,
    )

    def do_verify(files: Iterable[RPMFile]) -> list[pathlib.Path]:
        """Verify some files."""
        res: Final = []
        for pkg in files:
            output = cfg.check_output(["rpm", "-Kv", "--", str(pkg.path)])
            if not any(re_sig.search(line) for line in output.splitlines()):
                res.append(pkg.path)

        return res

    return do_verify([built.source]) + do_verify(built.binaries)
