# SPDX-FileCopyrightText: StorPool <support@storpool.com>
# SPDX-License-Identifier: BSD-2-Clause
"""Scan a Yum repository's metadata."""

from __future__ import annotations

import dataclasses
import gzip
import pathlib
import re
from typing import IO, TYPE_CHECKING, cast

from defusedxml import ElementTree as DefusedElementTree

from . import rpm as s_rpm


if TYPE_CHECKING:
    from typing import Final
    from xml.etree import ElementTree  # noqa: S405  # we lie to mypy about that one


ARCH_VALID: Final = (
    "x86_64",
    "noarch",
)

REPOMD_BASE: Final = "http://linux.duke.edu/metadata"

NS_COMMON: Final = f"{REPOMD_BASE}/common"
NS_FILELISTS: Final = f"{REPOMD_BASE}/filelists"
NS_OTHER: Final = f"{REPOMD_BASE}/other"
NS_REPO: Final = f"{REPOMD_BASE}/repo"

BNS_REPO: Final = f"{{{NS_REPO}}}"
BNS_RPM: Final = f"{{{NS_REPO}/rpm}}"

RE_NAMESPACED: Final = re.compile(
    r""" ^
    [{] (?P<namespace> [^}]+ ) [}]
    (?P<name> [^{}]+ )
    $""",
    re.X,
)


@dataclasses.dataclass(frozen=True)
class RepoListPackage:
    """A single package found in a repository."""

    path: pathlib.Path
    name: str
    arch: str
    source: str | None


@dataclasses.dataclass(frozen=True)
class RepoListData:
    """XML data scanned from a repository."""

    tag: str
    namespace: str
    root: ElementTree.Element


def cbyname(elem: ElementTree.Element, name: str) -> ElementTree.Element | None:
    """Return the first child by the given name or None."""
    for child in elem:
        data = RE_NAMESPACED.match(child.tag)
        if data is None:
            if child.tag == name:
                return child
        elif data.group("name") == name:
            return child

    return None


def cbyname_d(elem: ElementTree.Element, name: str, msg: str) -> ElementTree.Element:
    """Return the first child by the given name."""
    value: Final = cbyname(elem, name)
    if value is None:
        raise s_rpm.RepoError(msg)
    return value


def ctext(elem: ElementTree.Element, name: str) -> str | None:
    """Return a child element's text or None."""
    child: Final = cbyname(elem, name)
    if child is None:
        return None
    return child.text


def ctext_d(elem: ElementTree.Element, name: str, msg: str) -> str:
    """Return a child element's text."""
    value: Final = ctext(elem, name)
    if value is None:
        raise s_rpm.RepoError(msg)
    return value


def cattr(elem: ElementTree.Element, name: str, attr: str) -> str | None:
    """Return a child element's attribute or None."""
    child: Final = cbyname(elem, name)
    if child is None:
        return None
    return child.attrib.get(attr)


def cattr_d(elem: ElementTree.Element, name: str, attr: str, msg: str) -> str:
    """Return a child element's attribute."""
    value: Final = cattr(elem, name, attr)
    if value is None:
        raise s_rpm.RepoError(msg)
    return value


@dataclasses.dataclass(frozen=True)
class RepoList:
    """Parse a single repository package list."""

    base: pathlib.Path
    path: pathlib.Path
    type: str = dataclasses.field(init=False)
    data: RepoListData = dataclasses.field(init=False)
    packages: dict[str, RepoListPackage] = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        """Store the base directory and the package list path."""
        object.__setattr__(self, "data", self.parse_data())
        object.__setattr__(self, "type", self.data.tag)
        object.__setattr__(self, "packages", self.parse_packages())

    def parse_data(self) -> RepoListData:
        """Get the root element of the package list."""
        with gzip.open(self.path.open(mode="rb"), mode="rb") as infile:
            # https://github.com/python/typeshed/issues/2580 ...nice!
            data: Final = DefusedElementTree.parse(cast(IO[bytes], infile)).getroot()

        nsdata: Final = RE_NAMESPACED.match(data.tag)
        if not nsdata:
            raise s_rpm.RepoError(f"No root tag namespace in {self.path}")
        nspace, tag = nsdata.group("namespace"), nsdata.group("name")
        if nspace not in {NS_COMMON, NS_FILELISTS, NS_OTHER}:
            raise s_rpm.RepoError(f"Unexpected root tag namespace '{nspace}' in {self.path}")

        return RepoListData(tag=tag, namespace=f"{{{nspace}}}", root=data)

    def parse_packages(self) -> dict[str, RepoListPackage]:
        """Parse the list of packages."""
        packages: Final = {}
        data: Final = self.data
        for child in data.root:
            if child.tag != f"{data.namespace}package":
                continue

            if self.type == "metadata":
                name = ctext_d(
                    child,
                    "name",
                    f"No package name in {self.path}",
                )
                arch = ctext_d(
                    child,
                    "arch",
                    f"No package arch for {name} in {self.path}",
                )
                path = self.base / cattr_d(
                    child,
                    "location",
                    "href",
                    f"No package location for {name} in {self.path}",
                )
            else:
                if "name" not in child.attrib:
                    raise s_rpm.RepoError(f"No package name in {self.path}")
                name = child.attrib["name"]
                if "arch" not in child.attrib:
                    raise s_rpm.RepoError(f"No package arch for {name} in {self.path}")
                arch = child.attrib["arch"]
                path = pathlib.Path()

            fmt = cbyname(child, "format")
            source = ctext(fmt, "sourcerpm") if fmt is not None else None

            packages[name] = RepoListPackage(path=path, name=name, arch=arch, source=source)

        return packages


@dataclasses.dataclass(frozen=True)
class RepoMD:
    """Parse a repository definition file."""

    base: pathlib.Path
    path: pathlib.Path
    primary: RepoList = dataclasses.field(init=False)
    filelists: RepoList = dataclasses.field(init=False)
    other: RepoList = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        """Parse the file and the lists that it describes."""
        self.parse()

    def parse_list(self, child: ElementTree.Element) -> RepoList:
        """Parse a single list element."""
        href: Final = cattr_d(
            child,
            "location",
            "href",
            f"No location[href] in {self.path}",
        )
        location: Final = self.base / href

        size_str: Final = ctext_d(child, "size", f"No size in {self.path}")
        size: Final = int(size_str)

        if location is None or size is None:
            raise s_rpm.RepoError(f"Incomplete list in {self.path}")
        if not location.is_file():
            raise s_rpm.RepoError(f"No {location}")
        actual_size: Final = location.stat().st_size
        if actual_size != size:
            raise s_rpm.RepoError(
                f"Size mismatch for {location}: expected {size}, got {actual_size}",
            )

        return RepoList(self.base, location)

    def parse(self) -> None:
        """Parse the file."""
        if not self.path.is_file():
            raise s_rpm.RepoError(f"Not a file: {self.path}")
        root: Final = DefusedElementTree.parse(str(self.path)).getroot()
        if root.tag != f"{BNS_REPO}repomd":
            raise s_rpm.RepoError(f"Unexpected repomd root element: {root.tag}")

        data: Final = {}
        for child in root:
            if child.tag != f"{BNS_REPO}data":
                continue
            ctype = child.attrib["type"]
            if ctype not in {"primary", "filelists", "other"}:
                continue
            data[ctype] = self.parse_list(child)

        if any(key not in data for key in ("primary", "filelists", "other")):
            raise s_rpm.RepoError(f"Incomplete repomd file: {self.path}")

        # Ruff doesn't understand frozen dataclasses yet
        object.__setattr__(self, "primary", data["primary"])  # noqa: PLC2801
        object.__setattr__(self, "filelists", data["filelists"])  # noqa: PLC2801
        object.__setattr__(self, "other", data["other"])  # noqa: PLC2801


@dataclasses.dataclass(frozen=True)
class RepoSubdir:
    """A subdirectory (sources, architecture, etc.) in a repository."""

    path: pathlib.Path
    path_repomd: pathlib.Path
    repomd: RepoMD


@dataclasses.dataclass(frozen=True)
class Repo:
    """Parse a repository's structure."""

    base: pathlib.Path
    sources: RepoSubdir = dataclasses.field(init=False)
    arches: dict[str, RepoSubdir] = dataclasses.field(init=False)

    def __post_init__(self) -> None:
        """Examine a repository at the specified path."""
        object.__setattr__(self, "sources", self.find_sources())
        object.__setattr__(self, "arches", self.find_arches())

    def check_repo_subdir(self, tag: str) -> RepoSubdir | None:
        """Check whether the specified subdir is a repository."""
        path: Final = self.base / tag
        path_repomd: Final = path / "repodata" / "repomd.xml"
        if not path_repomd.is_file():
            return None
        return RepoSubdir(
            path=path,
            path_repomd=path_repomd,
            repomd=RepoMD(path, path_repomd),
        )

    def find_sources(self) -> RepoSubdir:
        """Find the source repo."""
        repo: Final = self.check_repo_subdir("SRPMS")
        if repo is None:
            raise s_rpm.RepoError(f"No source repo at {self.base}")
        return repo

    def find_arches(self) -> dict[str, RepoSubdir]:
        """Find the architecture subdirectories."""
        found: Final = {
            arch: repo
            for arch, repo in (
                (candidate, self.check_repo_subdir(candidate)) for candidate in ARCH_VALID
            )
            if repo
        }
        if not found:
            raise s_rpm.RepoError(f"No architecture repos at {self.base}")
        return found

    def __str__(self) -> str:
        """Provide a human-readable representation."""
        return f"Repo({self.base} [{' '.join(sorted(self.arches))}])"
