# SPDX-FileCopyrightText: StorPool <support@storpool.com>
# SPDX-License-Identifier: BSD-2-Clause
"""Utility functions for the StorPool package helper."""

from __future__ import annotations

import dataclasses
import functools
import logging
import pathlib
import re
import sys
import typing

from sp_variant import variant

from . import defs


if typing.TYPE_CHECKING:
    from typing import Final


RE_CKSUM: Final = re.compile(
    r""" ^
    (?P<cksum> [0-9a-f]+ )
    [ ]
    [ *]
    (?P<name> \S .* )
    $ """,
    re.X,
)

RE_LIST_TREE: Final = re.compile(
    r""" ^
    (?P<mode> [0-7]+ ) \s+
    (?P<objtype> \S+ ) \s+
    (?P<hash> [0-9a-f]+ ) \s+
    (?P<path> .+ )
    $ """,
    re.X,
)


def list_files(cfg: defs.Config) -> list[pathlib.Path]:
    """List all the filenames for the specified branch."""

    def single(line: str) -> pathlib.Path:
        """Handle a single line of git ls-tree --full-tree output."""
        match: Final = RE_LIST_TREE.match(line)
        if not match:
            sys.exit(f"git ls-tree returned an unexpected line: {line!r}")
        return pathlib.Path(match.group("path"))

    return [
        single(line)
        for line in cfg.check_output(
            ["git", "ls-tree", "--full-tree", "-r", "-z", "HEAD"],
            cwd=cfg.path.project,
        )
        .rstrip("\0")
        .split("\0")
    ]


def list_packages_for_branch(  # noqa: C901  # This needs to be refactored
    cfg: defs.Config,
    all_files: list[pathlib.Path],
) -> dict[str, defs.Package]:
    """Get the checksum files for the specified Git branch."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    cfg.log.debug("list packages for the %(branch)r branch", {"branch": cfg.branch})

    var_all: Final = variant.get_all_variants()
    var_branch: Final = {var.builder.branch: var for var in var_all.values() if var.builder.branch}
    if cfg.branch.name not in var_branch:
        sys.exit(f"Expected to find {cfg.branch.name!r} in {var_branch.keys()!r}")

    res: Final[dict[str, defs.Package]] = {}
    for path in all_files:
        parts = path.parts
        name = parts[-1]
        if len(parts) != 3 or name != "sp-checksums.txt":  # noqa: PLR2004
            continue
        pkgname, distro = parts[0], parts[1]
        if distro not in defs.DISTROS:
            sys.exit(f"defs.Package {name!r}: unknown distro {distro!r}")

        pkg = res.get(pkgname)
        if pkg is None:
            pkg = defs.Package(name=pkgname, branch=cfg.branch.name, checksums={}, overlays=[])
            res[pkgname] = pkg
            cfg.log.debug(
                "- created a new %(pkgname)s package: %(pkg)s",
                {"pkgname": pkgname, "pkg": pkg},
            )
        else:
            cfg.log.debug(
                "- found a %(pkgname)S package: %(pkg)s",
                {"pkgname": pkgname, "pkg": pkg},
            )

        if distro in pkg.checksums:
            sys.exit(
                f"defs.Package {pkgname!r}: duplicate checksums file for "
                f"{distro!r}: first {pkg.checksums[distro]} now {path}",
            )
        pkg.checksums[distro] = path

        check_branch = cfg.branch.name
        overlay_base = cfg.path.project / pkgname / cfg.branch.flavor / "overlays"
        while check_branch:
            var = var_branch[check_branch]
            overlay_path = overlay_base / check_branch
            cfg.log.debug("    - checking for %(overlay_path)s", {"overlay_path": overlay_path})
            if overlay_path.is_dir():
                cfg.log.debug("      - found!")
                if overlay_path in pkg.overlays:
                    cfg.log.debug("      - already there")
                else:
                    pkg.overlays.insert(0, overlay_path)

            if not var.parent:
                break
            var = var_all[var.parent]
            check_branch = var.builder.branch

        cfg.log.debug("  - updated: %(pkg)s", {"pkg": pkg})

    return res


def list_packages(cfg: defs.Config) -> dict[str, defs.Package]:
    """Get a list of the packages in a branch that has them all."""
    branch: Final = defs.BRANCHES["centos/8"]
    all_files: Final = list_files(cfg)
    return list_packages_for_branch(dataclasses.replace(cfg, branch=branch), all_files)


def parse_checksums(cksums: pathlib.Path) -> dict[str, str]:
    """Parse a checksums file."""
    lines: Final = cksums.read_text(encoding="UTF-8").splitlines()
    if lines == ["native"]:
        return {}

    res: Final = {}
    for line in lines:
        match = RE_CKSUM.match(line)
        if not match:
            sys.exit(f"Unexpected checksum line in {cksums}: {line!r}")
        res[match.group("name")] = match.group("cksum")

    return res


def get_package_for_branch(cfg: defs.Config) -> defs.Package:
    """Ensure a package definition exists for the specified branch."""
    assert cfg.branch is not None  # noqa: S101  # mypy needs this
    assert cfg.pkgname is not None  # noqa: S101  # mypy needs this

    all_files: Final = list_files(cfg)
    all_packages: Final = list_packages_for_branch(cfg, all_files)
    package: Final = all_packages.get(cfg.pkgname)
    if package is None:
        sys.exit(f"Unknown package {cfg.pkgname}")

    return package


@functools.lru_cache
def build_logger(*, debug: bool = False, verbose: bool = False) -> logging.Logger:
    """Build a logger that outputs to the standard output and error streams.

    Messages of level `WARNING` and higher go to the standard error stream.
    If `verbose` is true, messages of level `INFO` go to the standard output stream.
    If `debug` is true, messages of level `DEBUG` also go to the standard error stream.
    """
    logger: Final = logging.getLogger("sp-pymod")
    logger.setLevel(logging.DEBUG if debug else logging.INFO if verbose else logging.WARNING)
    logger.propagate = False

    diag_handler: Final = logging.StreamHandler(sys.stderr)
    diag_handler.setLevel(logging.DEBUG if debug else logging.WARNING)
    diag_handler.addFilter(lambda rec: rec.levelno != logging.INFO)
    logger.addHandler(diag_handler)

    if verbose:
        info_handler: Final = logging.StreamHandler(sys.stdout)
        info_handler.setLevel(logging.INFO)
        info_handler.addFilter(lambda rec: rec.levelno == logging.INFO)
        logger.addHandler(info_handler)

    return logger
